package com.hello2morrow.sonargraph.languageprovider.cplusplus.controller.settings

import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.hello2morrow.foundation.file.FileUtility
import com.hello2morrow.foundation.utilities.ExceptionUtility
import com.hello2morrow.foundation.utilities.StringUtility
import com.hello2morrow.sonargraph.core.controller.system.MSBuildException
import com.hello2morrow.sonargraph.core.controller.system.MSBuildExecutor
import com.hello2morrow.sonargraph.languageprovider.cplusplus.model.path.CPlusPlusFileType
import com.hello2morrow.sonargraph.languageprovider.cplusplus.model.settings.SourceFileBean

import de.schlichtherle.truezip.file.TFile


class CPlusPlusMSBuildExecutor extends MSBuildExecutor
{
    private static final Logger LOGGER = LoggerFactory.getLogger(CPlusPlusMSBuildExecutor.class)
    private final TFile m_installationDirectory;
    private final String m_architecture;
    private final String m_toolsVersion;

    private static final String COMPILE_FILES_PREFIX = "Compile files: "
    private static final String EXCLUDED_FILES_PREFIX = "Excluded files: "
    private static final String MACROS_PREFIX = "Macros: "
    private static final String INCLUDE_DIRS_PREFIX = "Include directories: "
    private static final String USING_DIR_PREFIX = "Using directories: "
    private static final String RUNTIME_PREFIX = "Runtime: "
    private static final String MANAGED_PREFIX = "Managed: "
    private static final String FORCED_INCLUDES_PREFIX = "Forced includes: "

    public CPlusPlusMSBuildExecutor(TFile installationDirectory, String architecture, TFile projectFile, String toolsVersion, TFile solutionDir)
    {
        super(projectFile, solutionDir);
        assert installationDirectory != null : "Parameter 'installationDirectory' must not be null";
        assert installationDirectory.isDirectory() : "Parameter 'installationDirectory' must be an existing direcotry";
        assert architecture != null && architecture.length() > 0 : "Parameter 'architecture' must not be empty";
        assert toolsVersion != null && toolsVersion.length() > 0 : "Parameter 'toolsVersion' must not be empty";

        m_installationDirectory = installationDirectory;
        m_architecture = architecture;
        m_toolsVersion = toolsVersion;
    }


    private void addIfNotPresent(List<String> list, String value)
    {
        if (!list.contains(value))
        {
            list.add(value);
        }
    }

    @Override
    protected Map<String, String> getProperties()
    {
        Map<String, String> props = super.getProperties();
        props.put("toolsVersion", m_toolsVersion);

        return props;
    }

    @Override
    protected String getTemplateName()
    {
        return "msbuild.template";
    }


    @Override
    protected String getWrapperFileName()
    {
        return "h2m" + CPlusPlusFileType.VISUAL_STUDIO_PROJECT_FILE.getDefaultExtension();
    }

    public List<SourceFileBean> runMsBuild(final String configuration, final String platform) throws MSBuildException
    {
        TFile wrapper = writeProjectWrapper()
        int result = -1;

        try
        {
            List<String> command = new ArrayList<String>()
            command.add(new TFile(m_installationDirectory, "vcvarsall.bat").getNormalizedAbsolutePath());
            command.add(m_architecture);
            command.add("&&");
            command.add("msbuild");
            command.add("/nologo");
            command.add("/t:h2moptions");
            command.add("/p:Configuration=$configuration")
            command.add("/p:Platform=$platform")

            if (getSolutionDir() != null)
            {
                String path = getSolutionDir().getNormalizedAbsolutePath();
                command.add("/p:SolutionDir=$path");
            }

            command.add(wrapper.getNormalizedAbsolutePath());

            LOGGER.info("Executing: {}", command);
            def writer = new StringWriter()
            try
            {
                def proc = command.execute()

                writer << proc.in
                writer << proc.err

                result = proc.waitFor()
            }
            catch(IOException ex)
            {
                String msg = "Failed to execute command: " + StringUtility.concat(command, " ");
                LOGGER.error(msg + "; exception: " + ExceptionUtility.collectFirstAndLast(ex));
                throw new MSBuildException(msg);
            }
            if (result != 0)
            {
                def msg = "Problem executing msbuild: $writer";
                LOGGER.error(msg);
                throw new MSBuildException(msg);
            }

            String files = StringUtility.EMPTY_STRING;
            String excludedFiles = StringUtility.EMPTY_STRING;
            String macros = StringUtility.EMPTY_STRING;
            String includes = StringUtility.EMPTY_STRING;
            String runtimes = StringUtility.EMPTY_STRING;
            String usingDirs = StringUtility.EMPTY_STRING;
            String managed = StringUtility.EMPTY_STRING;
            String forcedIncludes = StringUtility.EMPTY_STRING;

            writer.toString().eachLine
            {
                String line = it.trim()
                if (line.startsWith(COMPILE_FILES_PREFIX))
                {
                    files = line.substring(COMPILE_FILES_PREFIX.length())
                }
                else if (line.startsWith(EXCLUDED_FILES_PREFIX))
                {
                    excludedFiles = removeTrailingSeparator(line.substring(EXCLUDED_FILES_PREFIX.length()), "|");
                }
                else if (line.startsWith(MACROS_PREFIX))
                {
                    macros = removeTrailingSeparator(line.substring(MACROS_PREFIX.length()), "|")
                }
                else if (line.startsWith(INCLUDE_DIRS_PREFIX))
                {
                    includes = removeTrailingSeparator(line.substring(INCLUDE_DIRS_PREFIX.length()), "|");
                }
                else if (line.startsWith(USING_DIR_PREFIX))
                {
                    usingDirs = removeTrailingSeparator(line.substring(USING_DIR_PREFIX.length()), "|");
                }
                else if (line.startsWith(RUNTIME_PREFIX))
                {
                    runtimes = line.substring(RUNTIME_PREFIX.length())
                }
                else if (line.startsWith(MANAGED_PREFIX))
                {
                    managed = line.substring(MANAGED_PREFIX.length())
                }
                else if (line.startsWith(FORCED_INCLUDES_PREFIX))
                {
                    forcedIncludes = removeTrailingSeparator(line.substring(FORCED_INCLUDES_PREFIX.length()), "|");
                }
            }

            String[] fileList = files.split(";", -1)
            String[] excludedFilesList = excludedFiles.split(";", -1);
            String[] macroList = macros.split("\\|", -1);
            String[] includeList = includes.split("\\|", -1);
            String[] usingList = usingDirs.split("\\|", -1)
            String[] runtimeList = runtimes.split(";", -1)
            String[] managedList = managed.split(";", -1)
            String[] forcedIncludeList = forcedIncludes.split("\\|", -1)

            try
            {
                assert excludedFilesList.length == 0 || excludedFilesList.length == fileList.length : "Size of excluded files does not match"
                assert macroList.length == 0 || macroList.length == fileList.length : "Size of macros files does not match"
                assert includeList.length == 0 || includeList.length == fileList.length : "Size of includes does not match"
                assert runtimeList.length == 0 || runtimeList.length == fileList.length : "Size of runtimes does not match"
                assert usingList.length == 0 || usingList.length == fileList.length : "Size of using does not match"
                assert forcedIncludeList.length == 0 || forcedIncludeList.length == fileList.length : "Size of forced includes does not match"
            }
            catch (AssertionError assertion)
            {
                result = -1;

                //Better debugging support:
                LOGGER.error("Failed to process MSBuild result: ", assertion);
                LOGGER.error("Files [{}]]: {}", fileList.length, files);
                LOGGER.error("Excluded files [{}]: {}", excludedFilesList.length, excludedFiles.toString());
                LOGGER.error("Macro [{}]: {}", macroList.length, macros);
                LOGGER.error("Includes [{}]: {}", includeList.length, includes);
                LOGGER.error("Using [{}]: {}", usingList.length, usingDirs);
                LOGGER.error("Runtimes [{}]: {}", runtimeList.length, runtimes);
                LOGGER.error("Managed [{}]: {}", managedList.length, managed);
                LOGGER.error("Forced includes [{}]: {}", forcedIncludeList.length, forcedIncludes);
                throw assertion;
            }

            List<SourceFileBean> beans = new ArrayList<>()
            List<String> lastOptions = null;

            for (int i = 0; i < fileList.length; i++)
            {
                String file = fileList[i]
                if (excludedFilesList.length > 0)
                {
                    String excluded = excludedFilesList[i].trim()
                    if (Boolean.parseBoolean(excluded))
                    {
                        LOGGER.info("Excluding file for '{}|{}' : {}", configuration, platform, file);
                        continue
                    }
                }

                List<String> options = new ArrayList<>()
                if (includeList.length > 0)
                {
                    boolean currentDirectoryFound = false;
                    for (String inc : includeList[i].split(";"))
                    {
                        if (inc.length() > 0)
                        {
                            String cleaned = inc.replace("\"", "");
                            if (cleaned.equals("."))
                            {
                                currentDirectoryFound = true;
                            }
                            options.add(new String("-I$cleaned"));
                        }
                    }
                    if (!currentDirectoryFound)
                    {
                        options.add("-I.");
                    }
                }
                if (usingList.length > 0)
                {
                    for (String usingDir : usingList[i].split(";"))
                    {
                        if (usingDir.length() > 0)
                        {
                            String cleaned = usingDir.replace("\"", "");
                            options.add(new String("--using_directory=$cleaned"))
                        }
                    }
                }

                if (forcedIncludeList.length > 0)
                {
                    for (String inc : forcedIncludeList[i].split(";"))
                    {
                        if (inc.length() > 0)
                        {
                            String cleaned = inc.replace("\"", "");
                            options.add(new String("--preinclude=$cleaned"))
                        }
                    }
                }

                if (macroList.length > 0)
                {
                    for (String macro : macroList[i].split(";"))
                    {
                        if (macro.length() > 0)
                        {
                            options.add(new String("-D$macro"))
                        }
                    }
                }

                String runtime = runtimeList[i];

                if (runtime == null)
                {
                    runtime = "";
                }
                else
                {
                    runtime = runtime.toLowerCase();
                }
                switch (runtimeList[i].toLowerCase())
                {
                    case "multithreadeddebug":
                        addIfNotPresent(options, "-D_MT")
                        addIfNotPresent(options, "-D_DEBUG")
                        break;
                    case "multithreaded":
                        addIfNotPresent(options, "-D_MT")
                        break;
                    case "multithreadeddll":
                        addIfNotPresent(options, "-D_MT")
                        addIfNotPresent(options, "-D_DLL")
                        break;
                    case "multithreadeddebugdll":
                        addIfNotPresent(options, "-D_MT")
                        addIfNotPresent(options, "-D_DEGUG")
                        addIfNotPresent(options, "-D_DLL")
                        break;
                    case "singlethreadeddebug":
                        addIfNotPresent(options, "-D_DEBUG")
                        break;
                    case "singlethreaded":
                        break;
                    case "singlethreadeddll":
                        addIfNotPresent(options, "-D_DLL")
                        break;
                    case "singlethreadeddebugdll":
                        addIfNotPresent(options, "-D_DEGUG")
                        addIfNotPresent(options, "-D_DLL")
                        break;
                }
                if (managedList.length == fileList.length && !managedList[i].equals("false"))
                {
                    options.add("--clr")
                }
                if (lastOptions == null)
                {
                    lastOptions = options;
                }
                else if (options.equals(lastOptions))
                {
                    options = lastOptions;
                }

                //we exclude stdafx.cpp because it is not a real source file, only used to create precompiled headers
                if (!file.toLowerCase().contains("stdafx.cpp"))
                {
                    if(!file.isEmpty())
                    {
                        beans.add(new SourceFileBean(file, options))
                    }
                }
            }
            return beans
        }
        finally
        {
            if (result != 0)
            {
                List<String> content = FileUtility.getFileContent(getProjectFile())
                LOGGER.error("Wrapped project file content: \n{}", StringUtility.concat(content, StringUtility.LINE_SEPARATOR));
            }
            try
            {
                wrapper.rm()
            }
            catch(IOException ex)
            {
                LOGGER.error("Failed to delete file '{}'", wrapper.getAbsolutePath());
            }
        }
    }

    private String removeTrailingSeparator(String input, String separator)
    {
        assert input != null && input.length() > 0 : "Input must not be empty"
        assert separator != null && separator.length() > 0 : "Separator must not be empty"

        if (input.endsWith(separator))
        {
            return input.substring(0, input.lastIndexOf(separator));
        }
        return input;

    }
}